import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import logging
import multiprocessing
from multiprocessing import Process
import signal
import threading
import time

from holon.HolonicAgent import HolonicAgent

import json
# from opencc import OpenCC

import helper
from holon.HolonicAgent import HolonicAgent

from abdi_config import AbdiConfig

logger = helper.get_logger()


class TestResp(HolonicAgent):
    def __init__(self, cfg):
        super().__init__(cfg)


    def on_connected(self):
        self.subscribe("test.send")
        
        
    def on_request(self, topic, payload):
        if "test.send" == topic:
            with open("D:\\Work\\Temp\\aaa.jpg", 'rb') as file:
                file_content = file.read()
            return "test.resp", file_content


    def on_message(self, topic:str, payload):
        pass


class TestSend(HolonicAgent):
    def __init__(self, cfg):
        super().__init__(cfg)


    def on_connected(self):
        self.subscribe("test.start")
        self.subscribe("test.resp", data_type="str", topic_handler=self.process_resp)
        # self.subscribe("doc.text.import", "str", self.handle_doc_text_import)
        pass


    def on_message(self, topic:str, payload):
        if "test.start" == topic:
            with open("D:\\Work\\Temp\\bbb.png", 'rb') as file:
                file_content = file.read()
            logger.debug(f"file_content: {len(file_content)}")
            self.request("test.send", file_content)
            
            
    def process_resp(self, topic, payload):
        # logger.info(f"TestSend1, Agent: {self.agent_id}, topic:{topic}, payload:{payload}")
        with open("D:\\Work\\Temp\\ccc.jpg", 'wb') as file:
            file.write(payload)


if __name__ == '__main__':
    print('***** Test start *****')

    def signal_handler(signal, frame):
        print("signal_handler")
        # exit(0)
    signal.signal(signal.SIGINT, signal_handler)

    multiprocessing.set_start_method('spawn')

    TestSend(AbdiConfig(helper.get_config())).start()
    # TestSend2(AbdiConfig(helper.get_config())).start()
    TestResp(AbdiConfig(helper.get_config())).start()

    print('***** Test STOP *****')



import pickle

with open("D:\\OneDrive\\圖片\\FXE86F3833C03D_000211.jpg", 'rb') as file:
    file_content = file.read()
print(f"file_content: {len(file_content)}")

# Example dictionary to serialize
data = {
    "name": "Alice", 
    "age": 30, 
    "city": "New York",
    "image": file_content
}

# Serializing the dictionary to bytes
serialized_data = pickle.dumps(data)

# Deserializing the bytes back to a dictionary
deserialized_data = pickle.loads(serialized_data)

# Displaying the results
# print("Serialized Data:", serialized_data)
# print("Deserialized Data:", deserialized_data)

print(f"image: {len(deserialized_data['image'])}")
with open("D:\\Temp\\xxx.jpg", 'wb') as file:
    file.write(deserialized_data["image"])



import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

import multiprocessing
import random
import signal
import time

import helper
from holon.HolonicAgent import HolonicAgent
from holon.logistics.loading_coordinator import LoadingCoordinator

from abdi_config import AbdiConfig

logger = helper.get_logger()


class TestLB(HolonicAgent):
    def __init__(self, cfg):
        super().__init__(cfg)


    def on_connected(self):
        logger.info(f"on_connected.")
        self.append_logistic(LoadingCoordinator(
            self,
            work_topic="test.work",
            work_handler=self.do_work,
            loading_evaluator=self.evaluate_loading))
        # self.subscribe("test.work", topic_handler=self.do_work)
        self.publish("test.work1")
        
        
    def evaluate_loading(self, topic, payload):
        loading_rate = random.randint(0, 100)
        # logger.info(f"loading_rate: {loading_rate}")
        
        return loading_rate
        

    def do_work(self, topic:str, payload):
        logger.info(f"{self.agent_id}({self.short_id})> Do work.")
        # time.sleep(.1)
        # logger.info(f"{self.agent_id}({self.short_id})> Do work done.")
        
        
class TestLB2(TestLB):
    def evaluate_loading(self, topic, payload):
        time.sleep(1.5)
        return super().evaluate_loading(topic, payload)


if __name__ == '__main__':
    print('***** Test loading balance *****')

    def signal_handler(signal, frame):
        print("signal_handler")
    signal.signal(signal.SIGINT, signal_handler)

    multiprocessing.set_start_method('spawn')

    for _ in range(32):
        TestLB(AbdiConfig(helper.get_config())).start()
    # TestLB2(AbdiConfig(helper.get_config())).start()



import logging

from abdi_config import LOGGER_NAME
from holon.logistics.base_logistic import BaseLogistic


logger = logging.getLogger(LOGGER_NAME)



class BrokerLogistic(BaseLogistic):
    def __init__(self):
        pass


    def pack(self, payload):
        return payload


    def unpack(self, payload):
        return payload



import logging
import uuid

from abdi_config import LOGGER_NAME
from holon.HolonicAgent import HolonicAgent
from holon.logistics.base_logistic import BaseLogistic
from holon.logistics.payload_wrapper import PayloadWrapper


logger = logging.getLogger(LOGGER_NAME)
PUBLISH_HEADER = "@response"
SUBSCRIBE_HEADER = "@request"


class ResponseLogistic(BaseLogistic):
    def __init__(self, agent:HolonicAgent):
        self.agent = agent
        self._payload_wrapper = PayloadWrapper(self.agent.agent_id)
        
        
    def publish(self, topic, payload):
        sender_id = self.request_payload['sender']
        request_id = self.request_payload['request_id']
        logistic_topic = f"{PUBLISH_HEADER}.{sender_id}.{request_id}.{topic}"
        packed_payload = self._payload_wrapper.wrap_for_response(payload, self.request_payload)
        # logistic_topic, packed_payload = self.pack(topic, payload)
        logger.debug(f"logistic_topic: {logistic_topic}, packed_payload: {str(packed_payload)[:300]}")
        self.agent.publish(logistic_topic, packed_payload)


    def subscribe(self, topic, topic_handler=None, datatype="str"):
        request_topic = f"{SUBSCRIBE_HEADER}.{topic}"
        logger.debug(f"request_topic: {request_topic}")
        self.agent.subscribe(request_topic, datatype, self.handle_request)
        
        if topic_handler:
            self.agent.set_topic_handler(topic, topic_handler)

        
    def handle_request(self, topic:str, payload):
        logger.debug(f"topic: {topic}, payload: {str(payload)[:300]}...")
        request_topic = topic[len(SUBSCRIBE_HEADER)+1:]
        self.request_payload = self._payload_wrapper.unpack(payload)
        content = self.request_payload["content"]
        # logger.debug(f"request_topic: {request_topic}, agent_id: {self.agent.agent_id}, content: {content}")
        self.agent._on_message(request_topic, content)


    # def pack(self, topic:str, payload):
    #     if topic == self.request_topic:
    #         packed = self._payload_wrapper.wrap_for_response(payload, self.request_payload)
    #         return f"{PUBLISH_HEADER}.{topic}", packed
    #     else:
    #         return topic, payload


    # def unpack(self, payload):
    #     unpacked = self._payload_wrapper.unpack(payload)
    #     return unpacked



import logging

import rospy
from std_msgs.msg import String, Int32, UInt8MultiArray

from broker.message_broker import MessageBroker
from broker.notifier import BrokerNotifier
from abdi_config import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class RosNoeticBroker(MessageBroker):
    def __init__(self, notifier:BrokerNotifier):
        self.pub = rospy.Publisher('chatter', String, queue_size=10)
        self.publishers = {}
        
        super().__init__(notifier=notifier)



    ###################################
    # Implementation of MessageBroker #
    ###################################
    
    
    def start(self, options:dict):
        rospy.init_node(self._notifier.name, anonymous=True)
       

    def stop(self):
        logger.info(f"Ros broker is stopping...")
        
        
    def _get_publisher(self, topic, data):
        type_name = type(data).__name__
        topic_key = f"{topic}_{type_name}"
        publisher = self.publishers.get(topic_key)
        if not publisher:
            if "str" == type_name:
                publisher = rospy.Publisher(topic, String, queue_size=10)
            elif "int" == type_name:
                publisher = rospy.Publisher(topic, Int32, queue_size=10)
            elif "bytes" == type_name:
                publisher = rospy.Publisher(topic, UInt8MultiArray, queue_size=10)
            else:
                raise TypeError(f"Unsupported data type: {type_name}")
            
            self.publishers[topic_key] = publisher
            
        return publisher


    def publish(self, topic:str, payload):
        logger.info(f"topic: {topic}, payload: {payload}")
        
        publisher = self._get_publisher(topic, payload)
        rospy.loginfo(payload)
        publisher.publish(payload)
        
    
    def _callback_with_topic(self, topic):
        def callback(data):
            self._notifier._on_message(topic, data)
            rospy.loginfo(f"Received on topic {topic}, data.data: {data.data}")

        return callback


    def subscribe(self, topic:str, data_type):
        logger.info(f"topic: {topic}, , data_type: {data_type}")

        if "str" == data_type:
            rospy.Subscriber(topic, String, self._callback_with_topic(topic))
        elif "int" == data_type:
            rospy.Subscriber(topic, Int32, self._callback_with_topic(topic))
        elif "bytes" == data_type:
            rospy.Subscriber(topic, UInt8MultiArray, self._callback_with_topic(topic))
        else:
            raise TypeError(f"Unsupported data type: {data_type}")



from multiprocessing import Process
import paho.mqtt.client as mqtt
from holon.HolonicAgent import HolonicAgent

class Heart(HolonicAgent):

    def __init__(self, cfg):
        super().__init__(cfg)

    # def _run(self):
    #     print(f"  run: {self.__class__.__name__}")
    #     HolonicAgent._terminate_lock.wait()
    #     print(f"  terminate: {self.__class__.__name__}")
        
    # def start(self):
    #     pass
        


import logging
import uuid

from abdi_config import LOGGER_NAME
from holon.HolonicAgent import HolonicAgent
from holon.logistics.base_logistic import BaseLogistic
from holon.logistics.payload_wrapper import PayloadWrapper


logger = logging.getLogger(LOGGER_NAME)
PUBLISH_HEADER = "@request"
SUBSCRIBE_HEADER = "@response"


class RequestLogistic(BaseLogistic):
    __handlers = {}
    
    
    def __init__(self, agent:HolonicAgent, request_id=""):
        self.agent = agent
        self.request_id = request_id
        self.response_topic_header = f"{SUBSCRIBE_HEADER}.{self.agent.agent_id}.{self.request_id}"
        logger.debug(f"self, agent_id: {self.agent.agent_id}, short_id: {self.agent.short_id}, request_id: {self.request_id}")
        self._payload_wrapper = PayloadWrapper(self.agent.agent_id)
        
        
    def publish(self, topic, payload):
        logistic_topic = f"{PUBLISH_HEADER}.{topic}"
        logger.debug(f"agent_id: {self.agent.agent_id}, request_id: {self.request_id}")
        packed_payload = self._payload_wrapper.wrap_for_request(payload, self.request_id)
        # logistic_topic, packed_payload = self.pack(topic, payload)
        logger.debug(f"logistic_topic: {logistic_topic}, packed_payload: {str(packed_payload)[:300]}")
        self.agent.publish(logistic_topic, packed_payload)


    def subscribe(self, topic, topic_handler=None, datatype="str"):
        response_topic = f"{self.response_topic_header}.{topic}"
        self.agent.subscribe(response_topic, datatype, self.handle_response)
        RequestLogistic.__handlers[self.response_topic_header] = topic_handler


    def handle_response(self, topic:str, payload):
        responsed_topic = topic[len(self.response_topic_header)+1:]
        unpacked = self._payload_wrapper.unpack(payload)
        logger.debug(f"topic: {topic}, unpacked: {str(unpacked)[:300]}")

        if topic_handler := RequestLogistic.__handlers[self.response_topic_header]:
            self.agent.set_topic_handler(responsed_topic, topic_handler)
        self.agent._on_message(responsed_topic, unpacked["content"])


    # def handle_response(self, topic:str, payload):
    #     responsed_topic = topic[len(self.response_topic_header)+1:]
    #     # unpacked = self.unpack(payload)
    #     unpacked = self._payload_wrapper.unpack(payload)
    #     if unpacked["receiver"] == self.agent.agent_id:
    #         if unpacked["request_id"] == self.request_id:
    #             logger.debug(f"MATCHED, agent_id: {self.agent.agent_id}, request_id: {self.request_id}")
    #             self.agent._on_message(responsed_topic, unpacked["content"])
    #         else:
    #             logger.debug(f"NOT matched, request_id: {unpacked['request_id']}, self.request_id: {self.request_id}")
    #     else:
    #         logger.debug(f"NOT matched, receiver: {unpacked['receiver']}, agent_id: {self.agent.agent_id}")
            


    # def pack(self, topic:str, payload):
    #     if topic == self.request_topic:
    #         # request_id = str(uuid.uuid4())
    #         packed = self._payload_wrapper.wrap_for_request(payload, self.request_id)
    #         # logger.debug(f"packed: {packed}")
    #         return f"{PUBLISH_HEADER}.{topic}", packed
    #     else:
    #         return topic, payload


    # def unpack(self, payload):
    #     unpacked = self._payload_wrapper.unpack(payload)
    #     return unpacked



import logging

from broker.message_broker import MessageBroker
from broker.notifier import BrokerNotifier
from abdi_config import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class EmptyBroker(MessageBroker):
    def __init__(self, notifier:BrokerNotifier):
        logger.info(f"Initialize...")
        
        super().__init__(notifier=notifier)



    ###################################
    # Implementation of MessageBroker #
    ###################################
    
    
    def start(self, options:dict):
        logger.info(f"Empty broker is starting. options:{options}")
       

    def stop(self):
        logger.info(f"Empty broker is stopping...")


    def publish(self, topic:str, payload):
        logger.info(f"topic: {topic}, payload: {payload}")
        
    
    def subscribe(self, topic:str, data_type):
        logger.info(f"topic: {topic}, data_type: {data_type}")



import json
import logging
from queue import Queue
import random
import threading
import time

from abdi_config import LOGGER_NAME
from holon.HolonicAgent import HolonicAgent
from holon.logistics.base_logistic import BaseLogistic


logger = logging.getLogger(LOGGER_NAME)
HEADER_RANKING = "@ranking"
HEADER_ELECTED = "@elected"


class LoadingCoordinator(BaseLogistic):
    def __init__(self, agent:HolonicAgent, loading_evaluator, datatype="str"):
        self.agent = agent
        self.topic_handler = None
        self.loading_evaluator = loading_evaluator
        self.loading_rate = 0
        self.candidates = None
        self.electing = False
        self.topic_payloads = Queue()

        
    def publish(self, topic, payload):
        self.agent.publish(topic, payload)


    def subscribe(self, topic, topic_handler=None, datatype="str"):
        self.topic_handler = topic_handler
        self.agent.subscribe(topic, datatype, self.start)
        self.agent.subscribe(f"{HEADER_RANKING}.{topic}", datatype, self.rank)
        self.agent.subscribe(f"{HEADER_ELECTED}.{topic}", datatype, self.elected)
        
        
    def reset(self):
        self._set_electing(False)
        self.candidates = []
        self.determine_delay = ThreadSafeCounter()
        
        
    def _set_electing(self, is_electing):
        self.electing = is_electing
        logger.debug(f"{self.agent.short_id}> set electing to {self.electing}")
        
        
    def start(self, topic:str, payload):
        logger.debug(f"{self.agent.short_id}> topic: {topic}, payload: {payload}")
        if self.electing:
            logger.warning(f"{self.agent.short_id}> electing")
            self.topic_payloads.put((topic, payload))
            return
        self.reset()
        self._set_electing(True)
        
        self.loading_rate = self.loading_evaluator(topic, payload)
        self.rank_number = self.loading_rate * 10000 + random.randint(0, 9999)

        rank_payload = {
            "agent_id": self.agent.agent_id,
            "rank_number": self.rank_number
        }
        self.candidates.append(rank_payload)
        self.agent.publish(f"{HEADER_RANKING}.{topic}", json.dumps(rank_payload))
        
        threading.Timer(.1, self.determine, args=(topic, payload)).start()


    def rank(self, topic:str, payload):
        if not self.electing:
            logger.warning(f"{self.agent.short_id}> NOT electing")
            return
            
        rank_payload = json.loads(payload.decode())
        if rank_payload['agent_id'] != self.agent.agent_id:
            self.candidates.append(rank_payload)
            
        self.determine_delay.add(0.01)
        
        
    def determine(self, topic:str, payload):
        if not self.electing:
            logger.warning(f"{self.agent.short_id}> NOT electing")
            return
        
        determine_delay = self.determine_delay.get_value()
        while determine_delay > 0:
            self.determine_delay.substract(determine_delay)
            time.sleep(determine_delay)
            determine_delay = self.determine_delay.get_value()
        
        logger.debug(f"{self.agent.short_id}> Candidates: {self.candidates}")
        # logger.debug(f"{self.agent.short_id}> candidates size: {len(self.candidates)}")
        if len(self.candidates):
            min_agent = min(self.candidates, key=lambda x: (x['rank_number'], x['agent_id']))
            if self.agent.agent_id == min_agent['agent_id']:
                logger.info(f"{self.agent.short_id}> Elected for topic: {topic}, payload: {payload}")
                self.agent.publish(f"{HEADER_ELECTED}.{topic}", self.agent.agent_id)
                
                if self.topic_handler:
                    self.topic_handler(topic, payload)
                else:
                    self.agent.on_message(topic, payload)
        
        
    def elected(self, topic:str, payload):
        self.reset()
        
        elected_agent_id = payload.decode()
        if not self.topic_payloads.empty():
            work = self.topic_payloads.get()
            if self.agent.agent_id == elected_agent_id:
                next_topic, next_payload = work
                logger.info(f"{self.agent.short_id}> Next work, topic: {next_topic}, payload: {next_payload}")
                self.agent.publish(next_topic, next_payload)
            # self.start(topic=work[0], payload=work[1])


    def pack(self, topic:str, payload):
        return topic, payload


    def unpack(self, payload):
        return payload
            
            
            
class ThreadSafeCounter:
    def __init__(self, initial=0):
        self.value = initial
        self.lock = threading.Lock()
    
    def add(self, number):
        with self.lock:
            self.value += number
    
    def substract(self, number):
        with self.lock:
            self.value -= number
    
    def increment(self):
        with self.lock:
            self.value += 1
    
    def decrement(self):
        with self.lock:
            self.value -= 1
    
    def get_value(self):
        with self.lock:
            return self.value


from broker import BrokerType
from broker.empty_broker import EmptyBroker
from broker.mqtt_broker import MqttBroker
from broker.redis_broker import RedisBroker
from broker.ros_broker import RosBroker
# from broker.ros_noetic_broker import RosNoeticBroker
from broker.notifier import BrokerNotifier


class BrokerMaker():
    def create_broker(self, broker_type:BrokerType, notifier:BrokerNotifier):
        if broker_type is BrokerType.Redis:
            return RedisBroker(notifier)
        elif broker_type is BrokerType.MQTT:
            return MqttBroker(notifier)
        elif broker_type is BrokerType.ROS:
            return RosBroker(notifier)
        # elif broker_type is BrokerType.ROS:
        #     return RosNoeticBroker(notifier)
        elif broker_type is BrokerType.Empty:
            return EmptyBroker(notifier)
        else:
           raise TypeError(f"Unsupported broker type: {type(broker_type).__name__}.") 
        


from abc import ABC, abstractmethod
from typing import Any


class BaseLogistic(ABC):
    @abstractmethod
    def publish(self, topic, payload):
        """
        Abstract method that packages the payload.
        """
        pass


    def subscribe(self, topic):
        """
        Abstract method that packages the payload.
        """
        pass


    # @abstractmethod
    # def pack(self, payload) -> (str, Any):
    #     """
    #     Abstract method that packages the payload.
    #     """
    #     pass


    # @abstractmethod
    # def unpack(self, payload):
    #     """
    #     Abstract method that unpacks the payload.
    #     """
    #     pass
    